@@ -17,6 +17,10 @@ $Date: 1993/03/03 00:21:20 $
 /*#include <sys/time.h> seems to not be present on SVR3 systems */
 /* and it's not used anyway as far as I can tell */
 
+#ifdef HAVE_SYS_WAIT_H
+#include <sys/wait.h>
+#endif
+
 #include <ctype.h>
 
 #include "tcl.h"
@@ -32,15 +36,31 @@ $Date: 1993/03/03 00:21:20 $
 #include "tclRegexp.h"
 #include "exp_regexp.h"
 
-/* tests for if we are running this using /dev/tty */
-#define real_tty_output(x) (exp_stdin_is_tty && ((x)==1) || ((x)==exp_dev_tty))
-#define real_tty_input(x) (exp_stdout_is_tty && ((x)==0) || ((x)==exp_dev_tty))
+/*
+ * tests if we are running this using a real tty
+ *
+ * these tests are currently only used to control what gets written to the
+ * logfile.  Note that removal of the test of "..._is_tty" means that stdin
+ * or stdout could be redirected and yet stdout would still be logged.
+ * However, it's not clear why anyone would use log_file when these are
+ * redirected in the first place.  On the other hand, it is reasonable to
+ * run expect as a daemon in which case, stdin/out do not appear to be
+ * ttys, yet it makes sense for them to be logged with log_file as if they
+ * were.
+ */
+#if 0
+#define real_tty_output(x) (exp_stdout_is_tty && (((x)==1) || ((x)==exp_dev_tty)))
+#define real_tty_input(x) (exp_stdin_is_tty && (((x)==0) || ((x)==exp_dev_tty)))
+#endif
+
+#define real_tty_output(x) (((x)==1) || ((x)==exp_dev_tty))
+#define real_tty_input(x)  (((x)==0) || ((x)==exp_dev_tty))
 
 #define new(x)	(x *)ckalloc(sizeof(x))
 
 struct action {
 	char *statement;
-	int tty_reset;		/* true, reset tty mode upon action */
+	int tty_reset;		/* if true, reset tty mode upon action */
 	int iread;		/* if true, reread indirects */
 	int iwrite;		/* if true, write spawn_id element */
 	int timestamp;		/* if true, generate timestamp */
@@ -321,14 +341,14 @@ int master;
 
 static int
 update_interact_fds(interp,fd_count,fd_to_input,fd_list,input_base,
-			do_indirect,close_count,real_tty_caller)
+			do_indirect,config_count,real_tty_caller)
 Tcl_Interp *interp;
 int *fd_count;
 struct input ***fd_to_input;	/* map from fd's to "struct input"s */
 int **fd_list;
 struct input *input_base;
 int do_indirect;		/* if true do indirects */
-int *close_count;
+int *config_count;
 int *real_tty_caller;
 {
 	struct input *inp;
@@ -338,7 +358,7 @@ int *real_tty_caller;
 
 	int real_tty = FALSE;
 
-	*close_count = exp_close_count;
+	*config_count = exp_configure_count;
 
 	count = 0;
 	for (inp = input_base;inp;inp=inp->next) {
@@ -412,8 +432,6 @@ int *real_tty_caller;
 	return TCL_OK;
 }
 
-static int inter_indirect_key = 0;
-
 /*ARGSUSED*/
 static char *
 inter_updateproc(clientData, interp, name1, name2, flags)
@@ -423,14 +441,14 @@ char *name1;		/* Name of variable. */
 char *name2;		/* Second part of variable name. */
 int flags;		/* Information about what happened. */
 {
-	inter_indirect_key++;
+	exp_configure_count++;
 	return 0;
 }
 			
 #define finish(x)	{ status = x; goto done; }
 
-static char interpreter_cmd[] = "interpreter";
 static char return_cmd[] = "return";
+static char interpreter_cmd[] = "interpreter";
 
 /*ARGSUSED*/
 int
@@ -479,8 +497,6 @@ char **argv;
 
 	char *replace_user_by_process = 0; /* for -u flag */
 
-	int close_count;	/* monitor exp_close_count */
-
 	struct input *input_base;
 #define input_user input_base
 	struct input *input_default;
@@ -499,7 +515,7 @@ char **argv;
 	struct keymap **end_km;
 
 	int key;
-	int indirect_key;
+	int configure_count;	/* monitor reconfigure events */
 
 	if ((argc == 2) && exp_one_arg_braced(argv[1])) {
 		return(exp_eval_with_one_arg(clientData,interp,argc,argv));
@@ -962,7 +978,7 @@ char **argv;
 	/* "finish()" to terminate this procedure                      */
 	/***************************************************************/
 
-	status = update_interact_fds(interp,&input_count,&fd_to_input,&fd_list,input_base,1,&close_count,&real_tty);
+	status = update_interact_fds(interp,&input_count,&fd_to_input,&fd_list,input_base,1,&configure_count,&real_tty);
 	if (status == TCL_ERROR) finish(TCL_ERROR);
 
 	if (real_tty) {
@@ -980,8 +996,8 @@ char **argv;
 
 	key = expect_key++;
 
-	/* declare ourselves "in sync" with external view of indirects */
-	indirect_key = inter_indirect_key;
+	/* declare ourselves "in sync" with external view of close/indirect */
+	configure_count = exp_configure_count;
 
 #ifndef SIMPLE_EVENT
 	/* loop waiting (in event handler) for input */
@@ -1029,21 +1045,18 @@ char **argv;
 			/* the count-down code is broken) */
 		}
 
-		/* update the world, if:
-		   indirects have been updated or
-		   any fds have closed
-		 */
-		if (indirect_key != inter_indirect_key
-		    || exp_close_count != close_count) {
+		/* update the world, if necessary */
+		if (configure_count != exp_configure_count) {
 			status = update_interact_fds(interp,&input_count,
 					&fd_to_input,&fd_list,input_base,1,
-					&close_count,&real_tty);
+					&configure_count,&real_tty);
 			if (status) finish(status);
 		}
 
 		rc = exp_get_next_event(interp,fd_list,input_count,&m,timeout,key);
-		if (rc == EXP_TCLERROR)
-			return(TCL_ERROR);
+		if (rc == EXP_TCLERROR) return(TCL_ERROR);
+
+		if (rc == EXP_RECONFIGURE) continue;
 
 		if (rc == EXP_TIMEOUT) {
 			if (timeout_simple) {
@@ -1103,7 +1116,7 @@ char **argv;
 
 				/* avoid another function call if possible */
 				if (debugfile || is_debugging) {
-					debuglog("spawn id %d sent {%s}\r\n",m,
+					debuglog("spawn id %d sent <%s>\r\n",m,
 						exp_printify(u->buffer + u->size - cc));
 				}
 				break;
@@ -1145,7 +1158,7 @@ char **argv;
 		/* put regexp result in variables */
 		if (km && km->re) {
 #define INTER_OUT "interact_out"
-#define out(var,val)  debuglog("expect: set %s(%s) {%s}\r\n",INTER_OUT,var, \
+#define out(var,val)  debuglog("expect: set %s(%s) \"%s\"\r\n",INTER_OUT,var, \
 						dprintify(val)); \
 		    Tcl_SetVar2(interp,INTER_OUT,var,val,0);
 
@@ -1456,7 +1469,7 @@ got_action:
 
 				/* avoid another function call if possible */
 				if (debugfile || is_debugging) {
-					debuglog("spawn id %d sent {%s}\r\n",m,
+					debuglog("spawn id %d sent <%s>\r\n",m,
 						exp_printify(u->buffer + u->size - cc));
 				}
 				break;
@@ -1493,7 +1506,7 @@ got_action:
 		/* put regexp result in variables */
 		if (km && km->re) {
 #define INTER_OUT "interact_out"
-#define out(i,val)  debuglog("expect: set %s(%s) {%s}\r\n",INTER_OUT,i, \
+#define out(i,val)  debuglog("expect: set %s(%s) \"%s\"\r\n",INTER_OUT,i, \
 						dprintify(val)); \
 		    Tcl_SetVar2(interp,INTER_OUT,i,val,0);
 
@@ -1665,7 +1678,6 @@ got_action:
 		}
 
 		if (action) {
-got_action_2:
 			te = inter_eval(interp,action,m);
 			switch (te) {
 			case TCL_BREAK:
@@ -1771,7 +1783,7 @@ got_action_2:
 
 				/* avoid another function call if possible */
 				if (debugfile || is_debugging) {
-					debuglog("spawn id %d sent {%s}\r\n",m,
+					debuglog("spawn id %d sent <%s>\r\n",m,
 						exp_printify(u->buffer + u->size - cc));
 				}
 				break;
@@ -1817,11 +1829,6 @@ got_action_2:
 
 		/* put regexp result in variables */
 		if (km && km->re) {
-#define INTER_OUT "interact_out"
-#define out(i,val)  debuglog("expect: set %s(%s) {%s}\r\n",INTER_OUT,i, \
-						dprintify(val)); \
-		    Tcl_SetVar2(interp,INTER_OUT,i,val,0);
-
 			char name[20], value[20];
 			regexp *re = km->re;
 			char match_char;/* place to hold char temporarily */
@@ -1995,7 +2002,6 @@ got_action_2:
 		}
 
 		if (action) {
-got_action_3:
 			change = (action && action->tty_reset);
 			if (change && tty_changed)
 				exp_tty_set(interp,&tty_old,was_raw,was_echo);
